;----------------------------------------------------------------------------
pro mge_fit_sectors_regularized, radius, angle, counts, eps, $
    QBOUNDS=qbounds1, SOL=sol, _EXTRA=ex
;
; This program is a wrapper for mge_fit_sectors procedure and it accepts all keyword of that
; program. One should look at the documentation of mge_fit_sectors for usage details.
; The wrapper implements the method described in Section 2.2.2 of Cappellari (2002, MNRAS, 333, 400)
; to "regularize" an MGE model by restricting the allowed range in qObs of the Gaussians until
; the fit becomes unacceptable. In this way one ensures that the permitted galaxy inclinations
; are not being artificialy restricted to a smaller range than allowed by the data.
; The detailed approach implemented here is the one used for the MGE fits of the galaxies in
; the Atlas3D project and described in Section 3.2 of Scott et al. (2013, MNRAS, 432, 1894).
;
; The intended usage of this wrapper is the following:
;   1. First perform a standard MGE fit with mge_fit_sectors;
;   2. Once all parameters are OK and the fit looks good, change the name mge_fit_sectors
;      in your script into mge_fit_sectors_regularized to cleanup the final solution.
;      This because this wrapper calls mge_fit_sectors repeatedly, taking much longer,
;      so it is not useful to run it until all input parameters are settled.
;
; VERSION HISTORY:
;   V1.0.0: Michele Cappellari, Oxford, 22 January 2013
;   V1.0.1: Fixed program stop when (qmin EQ qmax). Thanks to Silvia Posacki (Bologna)
;       for reporting the problem and the solution. MC, Oxford, 9 May 2013
;   V1.0.2: Removed truncation of input eps in mge_fit_sectors.
;       MC, Atlantic Ocean, 28 March 2015
;   V1.0.3: Slight change to the case qmin==qmax. MC, Oxford, 30 May 2015
;   V1.0.4: Backported changes from Python version. MC, Oxford, 21 November 2017

if n_elements(qbounds1) eq 0 then begin
    qmin = 0d
    qmax = 1d
endif else begin
    qmin = qbounds1[0]
    qmax = qbounds1[1]
endelse

nq = ceil((qmax - qmin)/0.05d) + 1 ; Adopt steps of at least 0.05 in qObs
qrange = cap_range(qmin, qmax, nq)
bestnorm = 1d30
frac = 1.1d ; Allowed fractional increase in ABSDEV
dq = (qrange[1] - qrange[0])/2d

for j=0,nq-1 do begin
    qmin = qrange[j]
    sol0 = 0
    mge_fit_sectors, radius, angle, counts, eps, $
        ABSDEV=absdev, QBOUNDS=[qmin, qmax], SOL=sol0, _STRICT_EXTRA=ex
    print, '(minloop) qbounds =', qmin, qmax
    print, '############################################'
    if absdev gt bestnorm*frac then begin
        jbest = j - 1
        qmin = qrange[jbest]
        break ; stops if error increases more than frac
    endif else begin
        jbest = j
        bestnorm <= absdev
        sol = sol0
    endelse
endfor

for k=nq-2,jbest,-1 do begin
    qmax = qrange[k]
    sol0 = 0
    mge_fit_sectors, radius, angle, counts, eps, $
        ABSDEV=absdev, QBOUNDS=[qmin, qmax], SOL=sol0, _STRICT_EXTRA=ex
    print, '(maxloop) qbounds =', qmin, qmax
    print, '############################################'
    if absdev gt bestnorm*frac then begin
        qmax = qrange[k+1]
        break ; stops if error increases more than frac
    endif else begin
        bestnorm <= absdev
        sol = sol0
    endelse
endfor

print, '############################################'
print, 'Final Regularized MGE Solution:'
print, '  Total_Counts  Sigma_Pixels      qObs'
print, '############################################'
print, sol
print, '############################################'
print, 'Final qbounds', qmin, qmax
print, '############################################'

end
;----------------------------------------------------------------------------
